import lab as B
import numpy as np
import pytest
import tensorflow as tf
import torch
import varz.autograd
import varz.tensorflow
import varz.torch
import wbml.out as out
from varz import Vars

from .util import approx, Value


class OutStream:
    """Mock the streams of `wbml.out`."""

    def __init__(self):
        self.output = ''

    def write(self, msg):
        self.output += msg

    def __enter__(self):
        self._orig_streams = list(out.streams)
        out.streams[:] = [self]
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        out.streams[:] = self._orig_streams


@pytest.fixture(params=[(np.float64, varz.autograd.minimise_l_bfgs_b, {}),
                        (torch.float64, varz.torch.minimise_l_bfgs_b, {}),
                        (tf.float64, varz.tensorflow.minimise_l_bfgs_b, {}),
                        (np.float64, varz.autograd.minimise_adam,
                         {'rate': 1e-1}),
                        (torch.float64, varz.torch.minimise_adam,
                         {'rate': 1e-1}),
                        (tf.float64, varz.tensorflow.minimise_adam,
                         {'rate': 1e-1})])
def minimise_method(request):
    yield request.param


def test_minimise(minimise_method):
    dtype, minimise, kw_args = minimise_method
    vs = Vars(dtype=dtype)

    # Initialise a variable that is not used.
    vs.get(name='other')

    # Define some objective.
    def f(vs_):
        return (-3 - vs_.pos(name='x', init=5.)) ** 2

    # Minimise it, until convergence.
    val_opt = minimise(f, vs, iters=10000, **kw_args)

    # Check for equality up to three digits.
    approx(val_opt, 9, digits=3)
    approx(vs['x'], 0, digits=3)


def test_minimise_disconnected_gradient(minimise_method):
    dtype, minimise, kw_args = minimise_method
    vs = Vars(dtype=dtype)
    vs.get(name='x')

    # Check that optimiser runs for objective that returns the constant zero.
    minimise(lambda v: B.cast(v.dtype, 0), vs, **kw_args)


def test_minimise_exception(minimise_method):
    dtype, minimise, kw_args = minimise_method
    vs = Vars(dtype=dtype)

    first_call = Value(True)

    # Define an objective that sometimes fails after the first call.
    def f(vs_):
        if first_call.val:
            first_call.val = False
        else:
            if np.random.rand() > .5:
                raise Exception('Fail!')
        return vs_.get(name='x', init=5.) ** 2

    # Check that the optimiser runs.
    minimise(f, vs, **kw_args)


def test_minimise_zero_calls(minimise_method):
    dtype, minimise, kw_args = minimise_method
    vs = Vars(dtype=dtype)

    calls = Value(0)

    def f(vs_):
        calls.val += 1
        return vs_.get(name='x', init=5.) ** 2

    # Check that running the optimiser for zero iterations only incurs a
    # single call.
    minimise(f, vs, iters=0, **kw_args)
    assert calls.val == 1


def test_minimise_trace(minimise_method):
    dtype, minimise, kw_args = minimise_method

    def f(vs_):
        return vs_.get(name='x') ** 2

    # Test that `trace=False` prints nothing.
    with OutStream() as stream:
        minimise(f, Vars(dtype=dtype), trace=False, **kw_args)
        assert stream.output == ''

    # Test that `trace=False` prints something.
    with OutStream() as stream:
        minimise(f, Vars(dtype=dtype), trace=True, **kw_args)
        assert stream.output != ''



import inspect
from abc import ABCMeta, abstractmethod
from functools import wraps

from .vars import Vars


class VarsProxy:
    """A proxy for a :class:`.vars.Vars` object.

    Args:
        vs (:class:`.vs.Vars`): Variable container to wrap.
    """

    def __init__(self, vs):
        self.vs = vs

    def __getattr__(self, item):
        return getattr(self.vs, item)


def is_vars(x):
    return isinstance(x, (Vars, VarsProxy))


class Sequential(VarsProxy):
    """A proxy for a :class:`.vars.Vars` object that automatically names
    unnamed variables sequentially.

    Args:
        vs (:class:`.vs.Vars`): Variable container to wrap.
        prefix (str): Prefix for names in the sequence.
    """

    def __init__(self, vs, prefix):
        VarsProxy.__init__(self, vs)
        self.prefix = prefix
        self.count = 0

    def _get_var(self, method, args, kw_args):
        if 'name' not in kw_args:
            kw_args['name'] = f'{self.prefix}{self.count}'
            self.count += 1
        return getattr(self.vs, method)(*args, **kw_args)

    def get(self, *args, **kw_args):
        return self._get_var('get', args, kw_args)

    def positive(self, *args, **kw_args):
        return self._get_var('positive', args, kw_args)

    def pos(self, *args, **kw_args):
        return self._get_vars('pos', *args, **kw_args)

    def bounded(self, *args, **kw_args):
        return self._get_var('bounded', args, kw_args)

    def bnd(self, *args, **kw_args):
        return self._get_var('bnd', *args, **kw_args)


def _to_sequential(x, prefix):
    """Convert an variable container to sequential, if it is one.

    Args:
        x (object): Object to convert.
        prefix (str): Prefix for names in the sequence.

    Returns:
        object: `x` converted to sequential if it is a variable container,
            otherwise just `x`.
    """
    return Sequential(x, prefix) if is_vars(x) else x


def _extract_prefix_and_f(prefix_or_f):
    if prefix_or_f is None:
        # Not used as decorator and prefix is left unspecified.
        prefix = ''
        f = None
    elif isinstance(prefix_or_f, str):
        # Not used as decorator and prefix is specified.
        prefix = prefix_or_f
        f = None
    else:
        # Used as a decorator.
        prefix = ''
        f = prefix_or_f

    return prefix, f


def sequential(prefix_or_f=None):
    """Decorator that generates variable names for unnamed variables
    sequentially."""
    prefix, f = _extract_prefix_and_f(prefix_or_f)

    def decorator(f_):
        @wraps(f_)
        def wrapped_f(*args, **kw_args):
            # Replace all variable containers with their sequential variants.
            args = [_to_sequential(x, prefix) for x in args]
            kw_args = {k: _to_sequential(v, prefix) for k, v in kw_args.items()}

            return f_(*args, **kw_args)

        return wrapped_f

    # Return the decorated function directly if a function was given.
    return decorator if f is None else decorator(f)


class VariableType(metaclass=ABCMeta):
    def __init__(self, *args, **kw_args):
        self.args = args
        self.kw_args = kw_args

    def _get_var(self, method, name, init):
        # Copy dict to not modify the original.
        kw_args = dict(self.kw_args)

        # Set initial value.
        if (
                init is not None and
                'init' in kw_args and
                kw_args['init'] is not None
        ):
            raise ValueError(f'Initial value doubly specified: '
                             f'"{init}" and "{kw_args["init"]}".')
        kw_args['init'] = init

        # Set name.
        if name in kw_args and kw_args['name'] is not None:
            raise ValueError(f'Name doubly specified: '
                             f'{name} and {kw_args["name"]}.')
        kw_args['name'] = name

        return method(*self.args, **kw_args)

    @abstractmethod
    def instantiate(self, vs, name, init):  # pragma: no cover
        pass


class Unbounded(VariableType):
    def instantiate(self, vs, name, init):
        return self._get_var(vs.get, name, init)


class Positive(VariableType):
    def instantiate(self, vs, name, init):
        return self._get_var(vs.positive, name, init)


class Bounded(VariableType):
    def instantiate(self, vs, name, init):
        return self._get_var(vs.bounded, name, init)


def parametrised(prefix_or_f=None):
    """Decorator to specify variables with types."""

    prefix, f = _extract_prefix_and_f(prefix_or_f)

    def decorator(f_):
        @wraps(f_)
        def wrapped_f(*args, **kw_args):
            signature = inspect.signature(f_)

            # Convert all arguments to keyword arguments, and store them here.
            filled_kwargs = {}

            # Look for variable container.
            values = args + tuple(kw_args.values())
            num_containers = sum([is_vars(x) for x in values])
            if num_containers == 0:
                raise ValueError('No variable container found.')
            elif num_containers > 1:
                raise ValueError('Multiple variable containers found.')
            else:
                # There is exactly only variable container. Find it.
                vs = [x for x in values if is_vars(x)][0]

            # Walk through the arguments.
            for name, parameter in signature.parameters.items():
                annotation = parameter.annotation

                # Instantiate uninstantiated variable types.
                if (isinstance(annotation, type) and
                        issubclass(annotation, VariableType)):
                    annotation = annotation()

                # Check whether the parameter is a variable that needs to be
                # extracted.
                if isinstance(annotation, VariableType):
                    # Check for default value.
                    if parameter.default is not parameter.empty:
                        init = parameter.default
                    else:
                        init = None

                    # Store the instantiated variable.
                    filled_kwargs[name] = \
                        annotation.instantiate(vs, prefix + name, init)
                else:
                    if len(args) > 0:
                        # Positional arguments left. Must be next one.
                        filled_kwargs[name] = args[0]
                        args = args[1:]
                    else:
                        # No position arguments left. Extract from keywords.
                        filled_kwargs[name] = kw_args[name]
                        del kw_args[name]

            # Ensure that everything is parsed.
            if len(args) > 0:
                raise ValueError(f'{len(args)} positional argument(s) not '
                                 f'parsed.')
            if len(kw_args) > 0:
                raise ValueError(f'{len(kw_args)} keyword argument(s) not '
                                 f'parsed: {", ".join(kw_args.keys())}.')

            return f_(**filled_kwargs)

        return wrapped_f

    # Return the decorated function directly if a function was given.
    return decorator if f is None else decorator(f)



import numpy as np
import pytest
import tensorflow as tf
import torch
from varz import Vars

from .util import allclose, approx


@pytest.fixture(params=[np.float64, torch.float64, tf.float64])
def dtype(request):
    yield request.param


def test_get_vars():
    vs = Vars(np.int)

    # This test also tests that `Vars.get_vars` always returns the collection
    # of variables in the right order. This is important for optimisation.

    # Initialise some variables.
    vs.get(1, name='a')
    vs.get(2, name='1/b')
    vs.get(3, name='2/c')
    vs.unbounded(4, name='2/d')  # Test alias.

    # Test getting all.
    assert vs.get_vars() == [1, 2, 3, 4]
    assert vs.get_vars(indices=True) == [0, 1, 2, 3]

    # Test that names must exist.
    with pytest.raises(ValueError):
        vs.get_vars('e')

    # Test some queries.
    assert vs.get_vars('a') == [1]
    assert vs.get_vars('a', '*/b') == [1, 2]
    assert vs.get_vars('*/c', 'a') == [1, 3]
    assert vs.get_vars('*/c', '*/b', 'a') == [1, 2, 3]

    assert vs.get_vars('a', indices=True) == [0]
    assert vs.get_vars('a', '*/b', indices=True) == [0, 1]
    assert vs.get_vars('*/c', 'a', indices=True) == [0, 2]
    assert vs.get_vars('*/c', '*/b', 'a', indices=True) == [0, 1, 2]

    # Test some more queries.
    assert vs.get_vars('1/*') == [2]
    assert vs.get_vars('2/*') == [3, 4]
    assert vs.get_vars('1/*', '2/*') == [2, 3, 4]

    assert vs.get_vars('1/*', indices=True) == [1]
    assert vs.get_vars('2/*', indices=True) == [2, 3]
    assert vs.get_vars('1/*', '2/*', indices=True) == [1, 2, 3]

    # Test even more queries.
    assert vs.get_vars('*/b', '1/*') == [2]
    assert vs.get_vars('a', '2/*') == [1, 3, 4]
    assert vs.get_vars('a', '2/d', '2/*') == [1, 3, 4]
    assert vs.get_vars('2/d', '2/c', 'a', '1/*') == [1, 2, 3, 4]
    assert vs.get_vars('1/*') == [2]
    assert vs.get_vars('2/*') == [3, 4]
    assert vs.get_vars('1/*', '2/*') == [2, 3, 4]

    assert vs.get_vars('*/b', '1/*', indices=True) == [1]
    assert vs.get_vars('a', '2/*', indices=True) == [0, 2, 3]
    assert vs.get_vars('a', '2/d', '2/*', indices=True) == [0, 2, 3]
    assert vs.get_vars('2/d', '2/c', 'a', '1/*', indices=True) == [0, 1, 2, 3]
    assert vs.get_vars('1/*', indices=True) == [1]
    assert vs.get_vars('2/*', indices=True) == [2, 3]
    assert vs.get_vars('1/*', '2/*', indices=True) == [1, 2, 3]


def test_get_vars_cache_clearing():
    vs = Vars(np.float32)
    vs.get(name='var_a')
    assert vs.get_vars('var_*', indices=True) == [0]
    vs.get(name='var_b')
    assert vs.get_vars('var_*', indices=True) == [0, 1]


def test_positive():
    vs = Vars(np.float64)
    for _ in range(10):
        assert vs.pos() >= 0
        assert vs.positive() >= 0


def test_bounded():
    vs = Vars(np.float64)
    for _ in range(10):
        assert 10 <= vs.bnd(lower=10, upper=11) <= 11
        assert 10 <= vs.bounded(lower=10, upper=11) <= 11


def test_get_set_vector(dtype):
    vs = Vars(dtype=dtype)

    # Test stacking a matrix and a vector.
    vs.get(shape=(2,), name='a', init=np.array([1, 2]))
    vs.get(shape=(2, 2), name='b', init=np.array([[3, 4], [5, 6]]))
    allclose(vs.get_vector('a', 'b'), [1, 2, 3, 4, 5, 6])

    # Test setting elements.
    vs.set_vector(np.array([6, 5, 4, 3, 2, 1]), 'a', 'b')
    allclose(vs['a'], np.array([6, 5]))
    allclose(vs['b'], np.array([[4, 3], [2, 1]]))

    # Test setting elements in a differentiable way. This should allow for
    # any values.
    vs.set_vector(np.array(['1', '2', '3', '4', '5', '6']), 'a', 'b',
                  differentiable=True)
    assert np.all(vs['a'] == ['1', '2'])
    assert np.all(vs['b'] == np.array([['3', '4'], ['5', '6']]))


def test_assignment(dtype):
    vs = Vars(dtype=dtype)

    # Generate some variables.
    vs.get(1., name='unbounded')
    vs.pos(2., name='positive')
    vs.bnd(3., lower=0, upper=10, name='bounded')

    # Check that they have the right values.
    allclose(1., vs['unbounded'])
    allclose(2., vs['positive'])
    allclose(3., vs['bounded'])

    # Assign some new values.
    vs.assign('unbounded', 4.)
    vs.assign('positive', 5.)
    vs.assign('bounded', 6.)

    # Again check that they have the right values.
    allclose(4., vs['unbounded'])
    allclose(5., vs['positive'])
    allclose(6., vs['bounded'])

    # Differentiably assign new values. This should allow for anything.
    vs.assign('unbounded', 'value', differentiable=True)
    assert vs['unbounded'] == 'value'


def test_copy_torch():
    vs = Vars(torch.float64)

    # Create a variable.
    vs.pos(1, name='a')

    # Initialise vector packer.
    vs.get_vector()

    # Make a normal and detached copy.
    vs_copy = vs.copy()
    vs_detached = vs.copy(detach=True)

    # Require gradients for both.
    vs.requires_grad(True)
    vs_detached.requires_grad(True)

    # Do a backward pass.
    (vs_detached['a'] ** 2).backward()

    # Check that values are equal, but gradients only computed for one.
    assert vs['a'] == 1
    assert vs.get_vars('a')[0].grad is None
    assert vs_detached['a'] == 1
    assert vs_detached.get_vars('a')[0].grad == 2

    # Check that copied fields are, in fact, copies, and that the vector packer
    # is also copied.
    del vs.transforms[:]
    del vs.inverse_transforms[:]
    del vs.vars[:]
    vs.index_by_name.clear()

    for vs_copied in [vs_copy, vs_detached]:
        assert len(vs_copied.transforms) > 0
        assert len(vs_copied.inverse_transforms) > 0
        assert len(vs_copied.vars) > 0
        assert len(vs_copied.index_by_name) > 0
        assert vs_copied.vector_packer != None


def test_requires_grad_detach_vars_torch():
    vs = Vars(torch.float64)
    vs.pos(1, name='a')

    # Test that gradients need to first be required.
    with pytest.raises(RuntimeError):
        (2 * vs['a']).backward()

    # Test that gradients can be required and are then computed.
    vs.requires_grad(True)
    (2 * vs['a']).backward()
    assert type(vs.vars[0].grad) != type(None)

    # Test that variables can be detached.
    vs.pos(1, name='b')
    result = 2 * vs['b']
    vs.detach()
    with pytest.raises(RuntimeError):
        result.backward()


def test_source():
    vs = Vars(np.float32, source=np.ones(10))

    assert vs.get() == 1.
    approx(vs.pos(shape=(5,)), np.exp(np.ones(5)))
    approx(vs.pos(), np.exp(1.))
    with pytest.raises(ValueError):
        vs.pos(shape=(5,))

    # Test that the source variables are casted to the right data type.

    vs = Vars(np.float32, source=np.array([1]))
    assert vs.get().dtype == np.float32

    vs = Vars(np.float64, source=np.array([1]))
    assert vs.get().dtype == np.float64



import lab as B
from numpy.testing import assert_allclose, assert_array_almost_equal
from plum import Dispatcher

__all__ = ['allclose', 'approx', 'Value']

_dispatch = Dispatcher()


class Value:
    def __init__(self, val):
        self.val = val


@_dispatch({B.Number, B.NP, list})
def _to_numpy(x):
    return x


@_dispatch({B.Torch, B.TF})
def _to_numpy(x):
    return x.numpy()


def approx(x, y, digits=7):
    x = _to_numpy(x)
    y = _to_numpy(y)
    assert_array_almost_equal(x, y, decimal=digits)


def allclose(x, y):
    x = _to_numpy(x)
    y = _to_numpy(y)
    assert_allclose(x, y)
